import email
import logging
import json
import smtplib
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from unittest import _log

from pyModules.utils import is_base64
import validators
from sendgrid.helpers.mail import Mail, Attachment, FileContent, FileName, FileType, Disposition
from sendgrid import SendGridAPIClient
from email.mime.multipart import MIMEMultipart


class GenericSMTPProvider:
    a_content = None
    a_name = None
    a_ftype = None
    has_attachment = False

    def __init__(self, app_config, email=None, subject=None, content=None):
        self.email = email # This is misleading, it is actually a list of recipient emails
        self.subject = subject
        self.content = content
        self.pwd = app_config['SMTP']['password']
        self.user = app_config['SMTP']['username']
        self.server = app_config['SMTP']['server']
        self.port = app_config['SMTP']['port']
        self.app_config = app_config
        for email in self.email:
            if not validators.email(email):
                raise ValueError(f'Found invalid Email in recipient list')
        if 'smtp.gmail.com' in self.server and len(self.pwd) != 16:
            raise ValueError(f'Google App password must be 16 characters')
        if not self.subject or type(self.subject) != str:
            raise TypeError("subject must be a string of length > 0")
        if not self.content or type(self.content) != str:
            raise TypeError("content must be a string of length > 0")

    def attachment(self, content, name, ftype):
        self.has_attachment = True
        self.a_content = content
        self.a_name = name
        self.a_ftype = ftype
        if not is_base64(self.a_content):
            raise ValueError('Attachment not Base 64')
        if type(self.a_name) != str:
            raise TypeError("Name must be a string")
        if type(self.a_ftype) != str:
            raise TypeError("filetype must be a string")

    def send(self):
        try:
            msg = MIMEMultipart()
            msg['Subject'] = self.subject
            msg['From'] = "noreply@sensoraccess.co.uk"
            msg['To'] = ', '.join( self.email )

            # Attach Body
            body = MIMEText(self.content, 'html')
            msg.attach(body)

            # Attach Attachment
            if self.has_attachment:
                part = MIMEBase('application', "octet-stream")
                part.set_payload(self.a_content)
                part.add_header('Content-Transfer-Encoding', 'base64')
                part.add_header('Content-Disposition', 'attachment; filename="%s"' % self.a_name)
                msg.attach(part)  # msg is an instance of MIMEMultipart()

            # Connect/Authenticate & send
            server = smtplib.SMTP("smtp.gmail.com", self.port)
            server.starttls()
            server.login(self.user, self.pwd)
            server.sendmail(msg['From'], msg['To'], msg.as_string())
            server.quit()

        except Exception as e:
            raise Exception("email fail" + str(e))

        return


class SendGridEmailProvider:
    a_content = None
    a_name = None
    a_ftype = None
    has_attachment = False

    def __init__(self, app_config, email=None, subject=None, content=None):
        self.email = email
        self.subject = subject
        self.content = content
        self.app_config = app_config
        for email in self.email:
            if not validators.email(email):
                raise ValueError(f'Invalid Email')
        if not self.subject or type(self.subject) != str:
            raise TypeError("subject must be a string of length > 0")
        if not self.content or type(self.content) != str:
            raise TypeError("content must be a string of length > 0")

    def attachment(self, content, name, ftype):
        self.has_attachment = True
        self.a_content = content
        self.a_name = name
        self.a_ftype = ftype
        if not is_base64(self.a_content):
            raise ValueError('Attachment not Base 64')
        if type(self.a_name) != str:
            raise TypeError("Name must be a string")
        if type(self.a_ftype) != str:
            raise TypeError("filetype must be a string")

    def send(self):

        message = Mail(
            from_email="noreply@sensoraccess.co.uk",
            to_emails=self.email,
            subject=self.subject,
            html_content=self.content
        )
        if self.has_attachment:
            message.attachment = Attachment(
                FileContent(self.a_content),
                FileName(self.a_name),
                FileType(self.a_ftype),
                Disposition("attachment")
            )

        sg = SendGridAPIClient(self.app_config['Sendgrid']['api_key'])
        try:
            response = sg.send(message)
        except Exception as e:
            raise Exception("email fail" + str(e))
        logging.log.info(f"Completed with status code: {response.status_code}")
        logging.log.debug(f"body: {response.body}")
        headers = str(response.headers).replace("\n", " | ")
        logging.log.debug(f"headers: {headers}")


class BadgeEmail:

    def __init__(self, app_config, email=None, subject=None, content=None):
        self.email_provider = GenericSMTPProvider(app_config, email, subject, content)

    def send(self):
        self.email_provider.send()

    def attachment(self, content, name, ftype):
        self.email_provider.attachment(content, name, ftype)


class VisitorEmail:
    def __init__(self, app_config, name=None, company=None, rid=None, photo=None, signature=None):
        self.name = name
        self.company = company
        if type(self.name) != str:
            raise TypeError("name must be a string")
        if type(self.company) != str:
            raise TypeError("company must be a string")
        email = app_config['Admin']['emails']
        subject = "Visitor Request Received"

        # Compose HTML Email Body
        if photo:
            photo_html = f'''<img alt="Visitor Photo" src="data:image/png;base64,{photo}" />'''
        else:
            photo_html = '''<p>No Photo Provided</p>'''

        if signature:
            signature_html = f'''<img alt="Visitor Signature" src="data:image/png;base64,{signature}" />'''
        else:
            signature_html = '''<p>No Signature Provided</p>'''

        content = f'''<!DOCTYPE html><html lang='en'> 
                        <p>Visitor request received for {self.name} from {self.company}.  
                            <a href='{app_config['Admin']['host']}/admin?rid={str(rid)}'>Click here to approve request</a>
                        </p>
                        {photo_html}
                        {signature_html}
                        </html>'''
        print(content)
        super().__init__(app_config, email=email, subject=subject, content=content)


# --------------------[ Test Script ]-------------------- #

if __name__ == '__main__':
    with open("../main_config_OLD.json", "r") as read_file:
        temp_config = json.load(read_file)
    logging.basicConfig(format='%(asctime)s [%(levelname)s: %(name)s]\t%(message)s', level=logging.DEBUG)
    _log.info(f"Starting test")

    test_v_email = VisitorEmail(temp_config,
                                name="TEST_EMAIL",
                                company="RID_BLANK",
                                rid=""
                                )
    test_v_email.send()
