import base64
import json
import os
from datetime import datetime
from pathlib import Path
import logging

import requests
import validators
from pyModules.GuardPoint.api_cards import CardsAPI
from enum import Enum

from pyModules.GuardPoint.api_exceptions import CardHoldersAPIError, UnauthorizedError, NotFoundError
from pyModules.bearerToken import BearerAuth
from pyModules.utils import is_base64, is_hex

REQUEST_TIMEOUT = 3
log = logging.getLogger(__name__)


class NameFilterMode(Enum):
    EITHER_OR = " or "
    EXACT = " and "


class CardHoldersAPI:
    # cmd_new_cardholder = json.loads('''{
    #     "cardholderIdNumber": null,
    #     "description": null,
    #     "firstName": null,
    #     "lastName": null,
    #     "photo": null,
    #     "pinCode": null,
    #     "fromDateValid": null,
    #     "toDateValid": null
    # }''')

    cmd_new_cardholder_with_detail_and_card = json.loads('''{    
        "firstName": null,
        "lastName": null,
        "cardholderIdNumber": null,
        "photo": null,
        "cardholderTypeUID": "22222222-2222-2222-2222-222222222222",
        "securityGroupUID": "aaf86a73-5a95-4f32-9208-6fe1ef3ff89b",
        "fromDateValid": null,
        "isFromDateActive": true,
        "toDateValid": null,
        "isToDateActive": true,
        "cardholderPersonalDetail": {
            "email": null,
            "company": null
        },
        "cards": [
            {
                "description": "Visitor Card",
                "cardCode": null
            }
        ]
    }''')

    # visitor_signature = "cF_StringField_1"
    # host_signature = "cF_StringField_2"
    # nda_agreed = "cF_BoolField_1"

    def __init__(self, app_config, token=''):
        if not validators.url(app_config['GP10']['base_url']):
            raise ValueError('Bad URL')
        self.cmd_new_cardholder_with_detail_and_card['photo'] = app_config['DEFAULT']['photo']
        self.url = app_config['GP10']['base_url'] + "/odata/API_Cardholders"
        self.custom_field_url = app_config['GP10']['base_url'] + "/odata/API_CardholderCustomizedFields"
        self.auth = BearerAuth(token)
        self.customfields = app_config['Custom_Fields']

    def rename_custom_fields(self, card_holder):
        # TODO - Make custom field names dynamic
        # response = requests.get(self.url + "?"
        #                                    "$select=name, cardholderCFName&"
        #                                    "$filter="
        #                                    f"name eq '{self.customfields['host_sig_field']}' or",
        #                                    f"name eq '{self.customfields['visitor_sig_field']}' or",
        #                                    f"name eq '{self.customfields['nda_sig_field']}'",
        #                         auth=self.auth, timeout=REQUEST_TIMEOUT)
        # if response.status_code == requests.codes.ok:
        #     fields = response.json()['value']
        if "cardholderCustomizedField" in card_holder:
            # Rename and re-position custom fields
            card_holder["visitorSignature"] = card_holder["cardholderCustomizedField"][self.customfields["visitor_sig_field"]]
            card_holder["hostSignature"] = card_holder["cardholderCustomizedField"][self.customfields["host_sig_field"]]
            card_holder["ndaAgreed"] = card_holder["cardholderCustomizedField"][self.customfields["nda_aggree_field"]]
            del card_holder["cardholderCustomizedField"]
        return card_holder

    def get_offsite_count(self):
        response = requests.get(self.url + "?"
                                           "$count=true&"
                                           "$top=0&"
                                           "$filter=insidearea/uid eq 00000000-0000-0000-0000-100000000001 or insidearea/uid eq 00000000-0000-0000-0000-100000000003",
                                auth=self.auth, timeout=REQUEST_TIMEOUT)
        if response.status_code == requests.codes.ok:
            json_data = response.json()
            return {"count": json_data['@odata.count']}
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to get Offsite card count. {response.status_code}")

    def get_onsite_count(self):
        response = requests.get(self.url + "?"
                                           "$count=true&"
                                           "$top=0&"
                                           "$filter=insidearea/uid ne 00000000-0000-0000-0000-100000000001 and insidearea/uid ne 00000000-0000-0000-0000-100000000003",
                                auth=self.auth, timeout=REQUEST_TIMEOUT)
        if response.status_code == requests.codes.ok:
            json_data = response.json()
            return {"count": json_data['@odata.count']}
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to get Onsite card count. {response.status_code}")

    def get_offsite_cardholders(self, offset=0, count=10):
        response = requests.get(self.url + "?"
                                           f"$top={count}&"
                                           f"$skip={offset}&"
                                           "$select=lastName, firstName, lastPassDate, uid&"
                                           "$expand=insideArea($select=name)&"
                                           "$filter=insidearea/uid eq 00000000-0000-0000-0000-100000000001 or insidearea/uid eq 00000000-0000-0000-0000-100000000003",
                                auth=self.auth, timeout=REQUEST_TIMEOUT)
        if response.status_code == requests.codes.ok:
            card_holder = response.json()
            return card_holder['value']
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to get Offsite cardholders. {response.status_code}")

    def get_onsite_cardholders(self, offset=0, count=10):
        response = requests.get(self.url + "?"
                                           f"$top={count}&"
                                           f"$skip={offset}&"
                                           "$select=lastName, firstName, lastPassDate, uid&"
                                           "$expand=insideArea($select=name)&"
                                           "$filter=insidearea/uid ne 00000000-0000-0000-0000-100000000001 and insidearea/uid ne 00000000-0000-0000-0000-100000000003",
                                auth=self.auth, timeout=REQUEST_TIMEOUT)
        if response.status_code == requests.codes.ok:
            card_holder = response.json()
            return card_holder['value']
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to get Onsite cardholders. {response.status_code}")

    def lookup_cardholder_by_email(self, email, from_date_valid_filter=True):
        if not validators.email(email):
            raise ValueError(f'Malformed Email {email}')

        cmd = "?$expand=" \
              "cardholderPersonalDetail($select=email, company)&" \
              "$select=uid," \
              "lastName," \
              "firstName," \
              "cardholderPersonalDetail, " \
              "fromDateValid, " \
              "isFromDateActive, " \
              "toDateValid, " \
              "isToDateActive" \
              "&$filter=CardholderPersonalDetail/email eq " \
              f"'{email}'"

        if from_date_valid_filter is not False:
            today = datetime.utcnow().strftime('%Y-%m-%dT00:00:00Z')
            cmd += f" and fromDateValid le {today} and toDateValid ge {today}"

        response = requests.get(self.url + cmd
                                , auth=self.auth, timeout=REQUEST_TIMEOUT)

        if response.status_code == requests.codes.ok:
            card_holder = response.json()
            card_holders_value = card_holder['value']
            if len(card_holders_value) == 0:
                raise NotFoundError(f'No Cardholder Found')
            if len(card_holders_value) == 1:
                card_holders_value[0] = self.rename_custom_fields(card_holders_value[0])
                return card_holders_value[0]
            if len(card_holders_value) > 1:
                raise ValueError(f"Found multiple cardholders.")
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to get Card Holders from GP Server. {response.status_code}")

    def update_card_holder_signatures(self, uid, v_sig, h_sig, nda):
        if not validators.uuid(uid):
            raise ValueError(f'Malformed UID {uid}')
        if not is_base64(v_sig):
            raise ValueError(f'Malformed Visitor Signature')
        if not is_base64(h_sig):
            raise ValueError(f'Malformed Host Signature')

        cmd_update_cardholder_sigs = {
            self.customfields["visitor_sig_field"]: v_sig,
            self.customfields["host_sig_field"]: h_sig,
            self.customfields["nda_aggree_field"]: nda
        }
        response = requests.patch(self.custom_field_url + f"({uid})",
                                  auth=self.auth,
                                  json=cmd_update_cardholder_sigs,
                                  timeout=REQUEST_TIMEOUT)

        if response.status_code == requests.codes['no_content']:
            return
        elif response.status_code == requests.codes['unprocessable_entity']:
            raise CardHoldersAPIError("BadRequest - Unprocessable")
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to update Card Holder Signatures on GP Server. {response.status_code}")

    def get_card_holder(self, uid):
        if not validators.uuid(uid):
            raise ValueError(f'Malformed UID {uid}')

        response = requests.get(self.url + "(" + uid + ")?"
                                                       "$expand="
                                                       "cardholderCustomizedField("
                                                       f"$select={self.customfields['visitor_sig_field']},{self.customfields['host_sig_field']},{self.customfields['nda_aggree_field']}),"
                                                       "cardholderType($select=typeName),"
                                                       "cards($select=cardCode),"
                                                       "cardholderPersonalDetail($select=email, company), "
                                                       "securityGroup($select=name)&"
                                                       "$select=uid,"
                                                       "visitor_signature,"
                                                       "host_signature,"
                                                       "lastName,"
                                                       "firstName,"
                                                       "cardholderIdNumber,"
                                                       "status,"
                                                       "fromDateValid,"
                                                       "isFromDateActive,"
                                                       "toDateValid,"
                                                       "isToDateActive,"
                                                       "photo,"
                                                       "cardholderType,"
                                                       "cards,"
                                                       "cardholderPersonalDetail,"
                                                       "securityGroup"
                                , auth=self.auth, timeout=REQUEST_TIMEOUT)

        if response.status_code == requests.codes.ok:
            card_holder = response.json()
            card_holders_value = card_holder['value']
            return self.rename_custom_fields(card_holders_value[0])
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to get Card Holders from GP Server. {response.status_code}")

    def get_card_holder_photo(self, uid):
        if not validators.uuid(uid):
            raise ValueError(f'Malformed UID {uid}')

        response = requests.get(self.url + "(" + uid + ")?"
                                                       "$select=photo"
                                , auth=self.auth, timeout=REQUEST_TIMEOUT)

        if response.status_code == requests.codes.ok:
            card_holder = response.json()
            card_holders_value = card_holder['value']
            return self.rename_custom_fields(card_holders_value[0])
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to get Card Holders from GP Server. {response.status_code}")

    @staticmethod
    def _compose_filter(searchPhrase, mode=NameFilterMode.EXACT):
        filter_str = "$filter=(cardholderType/typeName eq 'Visitor')"
        if searchPhrase:
            words = list(filter(None, searchPhrase.split(" ")))[:5]  # Split by space, remove empty elements, ignore > 5 elements
            fields = ["firstName", "lastName", "CardholderPersonalDetail/company"]
            phrases = []
            for f in fields:
                for v in words:
                    phrases.append(f"contains({f}, '{v}')")
            filter_str += f" and ({' or '.join(phrases)})"
        filter_str += "&"
        return filter_str

    def get_card_holders(self, offset=0, limit=10, searchPhrase=None, mode=NameFilterMode.EXACT):
        filter_str = self._compose_filter(searchPhrase=searchPhrase, mode=mode)
        request_str = (self.url + "?" + filter_str +
                       "$expand="
                       "cardholderCustomizedField("
                       f"$select={self.customfields['visitor_sig_field']},{self.customfields['host_sig_field']},{self.customfields['nda_aggree_field']}),"
                       "cardholderType($select=typeName),"
                       "cards($select=cardCode),"
                       "cardholderPersonalDetail($select=email, company),"
                       "securityGroup($select=name)&"
                       "$select=uid,"
                       "lastName,"
                       "firstName,"
                       "cardholderIdNumber,"
                       "status,"
                       "fromDateValid,"
                       "isFromDateActive,"
                       "toDateValid,"
                       "isToDateActive,"
                       "photo,"
                       "cardholderType,"
                       "cards,"
                       "cardholderPersonalDetail,"
                       "securityGroup&"
                       "$orderby=fromDateValid desc&"
                       "$top=" + str(limit) + "&$skip=" + str(offset)
                       )
        log.debug("API Request string: " + request_str)
        response = requests.get(request_str, auth=self.auth, timeout=REQUEST_TIMEOUT)
        if response.status_code == requests.codes.ok:
            card_holders = response.json()
            card_holders = card_holders['value']
            for card_holder in card_holders:
                card_holder = self.rename_custom_fields(card_holder)
            return card_holders
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to get Card Holders from GP Server. {response.status_code}")

    # Returns newly created Cardholder's UID on Success
    # def create_card_holder(self, first_name='', last_name='', from_date_valid='', to_date_valid='',
    #                        photo=None):
    #
    #     self.cmd_new_cardholder["firstName"] = first_name
    #     self.cmd_new_cardholder["lastName"] = last_name
    #     self.cmd_new_cardholder["fromDateValid"] = from_date_valid + "Z"  # TODO: GetUsers Timezone
    #     self.cmd_new_cardholder["toDateValid"] = to_date_valid + "Z"
    #     self.cmd_new_cardholder["photo"] = photo
    #
    #     response = requests.post(self.url, auth=self.auth, json=self.cmd_new_cardholder)
    #
    #     if response.status_code == requests.codes['created']:
    #         new_card_holder = response.json()
    #         return new_card_holder['uid']
    #     elif response.status_code == requests.codes['unprocessable_entity']:
    #         raise CardHoldersAPIError("BadRequest - Cardholder Already Exists")
    #     else:
    #         print(response.status_code)
    #         raise CardHoldersAPIError("Failed to create new Card Holder on GP Server")

    # Returns newly created Cardholder's UID on Success
    def create_card_holder_with_detail_and_card(self, first_name='', last_name='', company='', from_datetime_valid='',
                                                to_datetime_valid='',
                                                photo=None, email=None, security_group_uid=None, card_code=None):

        self.cmd_new_cardholder_with_detail_and_card["firstName"] = first_name
        self.cmd_new_cardholder_with_detail_and_card["lastName"] = last_name
        self.cmd_new_cardholder_with_detail_and_card["fromDateValid"] = from_datetime_valid.replace('+00:00', 'Z')
        self.cmd_new_cardholder_with_detail_and_card["toDateValid"] = to_datetime_valid.replace('+00:00', 'Z')
        if photo:
            self.cmd_new_cardholder_with_detail_and_card["photo"] = photo  # TODO Move to js cardholder creation
        self.cmd_new_cardholder_with_detail_and_card["cardholderPersonalDetail"]["email"] = email
        self.cmd_new_cardholder_with_detail_and_card["cardholderPersonalDetail"]["company"] = company

        if card_code is None:
            self.cmd_new_cardholder_with_detail_and_card['cards'][0]['cardCode'] = CardsAPI.generate_card_code()
        else:
            if len(card_code) != 8:
                raise CardHoldersAPIError("BadRequest - Cardcode Length")
            if not is_hex(card_code):
                raise CardHoldersAPIError("BadRequest - Cardcode contains none hex characters")

            self.cmd_new_cardholder_with_detail_and_card['cards'][0]['cardCode'] = card_code

        if security_group_uid:
            if validators.uuid(security_group_uid):
                self.cmd_new_cardholder_with_detail_and_card['securityGroupUID'] = security_group_uid

        response = requests.post(self.url, auth=self.auth, json=self.cmd_new_cardholder_with_detail_and_card,
                                 timeout=REQUEST_TIMEOUT)

        if response.status_code == requests.codes['created']:
            new_card_holder = response.json()
            return new_card_holder['uid']
        elif response.status_code == requests.codes['unprocessable_entity']:
            raise CardHoldersAPIError("BadRequest - Cardholder Already Exists")
        elif response.status_code == requests.codes.unauthorized:
            raise UnauthorizedError(f"Token Rejected from GP Server. {response.status_code}")
        elif response.status_code == requests.codes['bad_request']:
            raise CardHoldersAPIError(f"GP Server rejected input. {response.status_code}")
        else:
            raise CardHoldersAPIError(f"Failed to create new Card Holder on GP Server. {response.status_code}")

    def delete_card_holder(self, uid):
        if not validators.uuid(uid):
            raise ValueError('BadRequest - uid')
        response = requests.delete(self.url + "(" + uid + ")", auth=self.auth, timeout=REQUEST_TIMEOUT)

        if response.status_code == requests.codes['no_content']:
            return True
        else:
            print(response.status_code)
            raise CardHoldersAPIError("Failed to delete Card Holder on GP Server")
