import logging
import os
import sqlite3
import uuid
from datetime import datetime
from sqlite3 import Error


class BaseDB:

    def __init__(self, db_path):
        """ create a database connection to a SQLite database """
        self.conn = None

        if os.path.isfile(db_path):
            print(f"db_path found: {db_path}")
        else:
            print(f"db_path not found: {db_path}")

        try:
            self.conn = sqlite3.connect(db_path)
            self.cursor = self.conn.cursor()
            print(sqlite3.version)
        except Error as e:
            print(e)

    def create(self, create_sql):
        self.conn.execute(create_sql)
        self.conn.commit()

    def insert(self, insert_sql):
        self.conn.execute(insert_sql)
        self.conn.commit()

    def update(self, update_sql):
        self.conn.execute(update_sql)
        self.conn.commit()

    def select(self, select_sql):
        self.cursor.execute(select_sql)
        return self.cursor.fetchall()

    def delete(self, delete_sql):
        self.cursor.execute(delete_sql)
        self.conn.commit()

    def __del__(self):
        self.conn.close()

    def __exit__(self):
        self.conn.close()


class RegisterDataBaseLocal(BaseDB):
    max_entries = 10

    def __init__(self, max_entries=10):
        self.max_entries = max_entries
        super().__init__("localDatabase.db")
        self.create_register_table()

    def create_register_table(self):  # This is a unique table, only used in a HealthPingDB

        self.create(r"""CREATE TABLE IF NOT EXISTS register (
                            uid BIGINT PRIMARY KEY,
                            timestamp VARCHAR(255), 
                            firstName VARCHAR(255),
                            lastName VARCHAR(255) NOT NULL,
                            company VARCHAR(255),
                            email VARCHAR(255),
                            photo VARCHAR(1000000),
                            signature VARCHAR(1000000)
                        ); """)

    def generate_request_uid(self):
        # DynamoDB attribute of type Number can store 126-bit integers
        # Shift right to ensure number can be represented with 64 bits
        uid = str(uuid.uuid1().int >> 64)

        if len(uid) > 18:
            uid = uid[0:18]

        return int(uid)

    def insert_visitor_request(self, firstname, lastname, company, email, photo, signature):
        assert isinstance(firstname, str) is True, "Visitor Requests must have a first name"
        assert isinstance(lastname, str) is True, "Visitor Requests must have a last name"

        if photo is None:
            photo = 'NULL'
        else:
            photo = f"'{photo}'"

        if signature is None:
            signature = 'NULL'
        else:
            signature = f"'{signature}'"

        # Crop Database
        self.crop_database(self.max_entries)

        # Insert new record
        uid = self.generate_request_uid()
        timestamp = int(datetime.now().timestamp())

        cmd = "INSERT INTO register (uid, timestamp, firstName, lastName, company, email, photo, signature) " \
              "VALUES ({uid},{timestamp},{firstname},{lastname},{company},{email},{photo},{signature});" \
            .format(uid=f"{uid}", timestamp=f"{timestamp}", firstname=f"'{firstname}'", lastname=f"'{lastname}'",
                    company=f"'{company}'", email=f"'{email}'", photo=photo, signature=signature)

        # logging.error("INSERT SQL: " + cmd)

        self.insert(cmd)

        return str(uid) + "-" + str(timestamp)

    def get_visitor_request(self, request_id):
        rid = request_id.split("-")
        uid = rid[0]

        rows = self.select(f"SELECT * FROM register WHERE uid = {uid};")

        return {
            "uid": rows[0][0],
            "timestamp": rows[0][1],
            "firstName": rows[0][2],
            "lastName": rows[0][3],
            "company": rows[0][4],
            "email": rows[0][5],
            "photo": rows[0][6],
            "signature": rows[0][7]
        }

    def crop_database(self, max_entries):
        assert isinstance(max_entries, int) is True, "max_entries must be a number"

        rows = self.select(f"SELECT COUNT(uid) FROM register;")
        count = rows[0][0]

        if count > max_entries:
            # Get oldest record
            rows = self.select(f"SELECT uid FROM register ORDER BY timestamp ASC LIMIT 1;")
            uid = rows[0][0]
            # delete record
            self.delete(f"DELETE FROM register WHERE uid='{uid}';")

        return

    def get_visitor_requests(self, offset=0, count=10):

        rows = self.select(f"SELECT * FROM register ORDER BY timestamp DESC;")

        return rows
