from datetime import datetime
import validators
import requests

from pyModules.handle_visitor_request import handle_new_visitor_request, handle_visitor_lookup, handle_get_visitor_requests
from pyModules.GuardPoint.api_cardholders import CardHoldersAPI, NameFilterMode
from pyModules.GuardPoint.api_exceptions import UnauthorizedError, ServiceUnavailable, NotFoundError
from pyModules.GuardPoint.api_securitygroups import SecurityGroupsAPI
from pyModules.GuardPoint.api_token import TokenAPIError
from pyModules.auth_token_mgr import AuthToken
from pyModules.sBadge import Badge
from pyModules.utils import is_base64, is_hex

MAX_PHOTO_SIZE = 900000


class SensorVMSError(Exception):

    def __init__(self, message="No Message defined", cause=""):
        self.message = message
        self.cause = cause
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message}'


accepted_operations = {
    # Public Operations
    'loadAdminPanel',
    'newVisitorRequest',
    'getToken',
    'emailSearchCardHolder',
    # Admin Operations (Requires a Token)
    'getVisitorRequest',
    'getVisitorRequests',
    'acceptVisitorRequest',
    'getCardHolders',
    'genCardHolderQR',
    'createCardHolder',
    'deleteCardHolder',
    'emailCardHolder',
    'getSecurityGroups'

}


def handler(event, context, app_config):
    try:
        token_auth = AuthToken(app_config)  # Class used to create and verify tokens
    except Exception as e:
        raise SensorVMSError("InternalServerError: " + str(e)) from e

    try:
        if "context" not in event:
            raise ServiceUnavailable('API Gateway Setup (Probable cause: Method request pass-through mapping template '
                                     'not set.)')

        if "resource-path" not in event['context']:
            raise ServiceUnavailable('resource-path not found')

        '''if event['context']['resource-path'] == "/":
            return home_html

        if event['context']['resource-path'] == "/register":
            if "rid" in event["params"]["querystring"]:
                return handle_visitor_lookup(event["params"]["querystring"]["rid"])
            return register_html

        if event['context']['resource-path'] == "/admin":
            return login_html'''

        if event['context']['resource-path'] == "/api":

            if 'body-json' not in event:
                raise ValueError("No JSON Parameters.")

            if "operation" not in event['body-json']:
                raise ValueError("Missing operation parameter.")

            operation = event['body-json']['operation']

            # Public (Non-Token) Operations
            if operation == 'getToken':
                if "username" not in event['body-json']:
                    raise ValueError('Missing Username')
                if "password" not in event['body-json']:
                    raise ValueError('Missing Password')

                jwt_token = token_auth.create(event['body-json']['username'], event['body-json']['password'])

                return {"token": jwt_token}

            # Check for Token
            if "token" not in event['body-json']:
                raise ValueError('Missing Token Key')
            if event['body-json']['token'] == "":
                raise UnauthorizedError('Empty Token')
            if event['body-json']['token'] is None:
                raise UnauthorizedError('Missing Token')

            # Verify the token, and extract the GP token, + check operation is permitted by token type (Admin/Guest)
            gp_token, token_type = token_auth.verify(event['body-json']['token'], operation)

            if operation == 'renewToken':
                jwt_token = token_auth.renew(gp_token, token_type)
                return {"token": jwt_token}

            if operation == 'newVisitorRequest':
                request_id = handle_new_visitor_request(event['body-json'], app_config)
                resp = {
                    "rid": request_id,
                }
                return resp

            # Admin(Token) Operations
            # Init GP API classes - Class CardHoldersAPI represents the GP-PRO API_CardHolders
            api = CardHoldersAPI(app_config, gp_token)
            api_security_groups = SecurityGroupsAPI(app_config, gp_token)

            # Muster
            if operation == 'getOnsiteCount':
                return api.get_onsite_count()
            if operation == 'getOffsiteCount':
                return api.get_offsite_count()
            if operation == 'getOnsiteCardholders':
                if "offset" not in event['body-json']:
                    raise ValueError('Missing Offset')
                if "count" not in event['body-json']:
                    raise ValueError('Missing Count')
                return api.get_onsite_cardholders(event['body-json']['offset'], event['body-json']['count'])
            if operation == 'getOffsiteCardholders':
                if "offset" not in event['body-json']:
                    raise ValueError('Missing Offset')
                if "count" not in event['body-json']:
                    raise ValueError('Missing Count')
                return api.get_offsite_cardholders(event['body-json']['offset'], event['body-json']['count'])

            # When the Visitor searches for themselves by email
            if operation == 'emailSearchCardHolder':
                if "email" not in event['body-json']:
                    raise ValueError('Missing Email')

                return api.lookup_cardholder_by_email(event['body-json']['email'])

            # The Visitor submits there signature images
            if operation == 'updateCardHolderSignatures':
                if "uid" not in event['body-json']:
                    raise ValueError('Missing UID')
                if "visitorSignature" not in event['body-json']:
                    raise ValueError('Missing visitorSignature')
                if "hostSignature" not in event['body-json']:
                    raise ValueError('Missing hostSignature')
                if "ndaAgreed" not in event['body-json']:
                    raise ValueError('Missing ndaAgreed')
                return api.update_card_holder_signatures(event['body-json']["uid"],
                                                         event['body-json']["visitorSignature"],
                                                         event['body-json']["hostSignature"],
                                                         event['body-json']["ndaAgreed"])

            # Get the details of a single Visitor Request
            if operation == 'getVisitorRequest':
                if "rid" not in event['body-json']:
                    raise ValueError('Missing RID')
                else:
                    return handle_visitor_lookup(app_config, event['body-json']["rid"])

            # Get a list of Visitor Requests
            if operation == 'getVisitorRequests':
                return handle_get_visitor_requests(app_config)

            # Load the Home Page
            '''if operation == 'loadAdminPanel':
                resp = {
                    "html": admin_html,
                }
                return resp'''

            if operation == 'emailCardHolder':
                if "uid" not in event['body-json']:
                    raise ValueError('Missing CardHolderUID')

                cardholder_data = api.get_card_holder(event['body-json']['uid'])

                Badge(cardholder_data, app_config).send()

                return

            # Generate QR from cardCode
            if operation == 'genCardHolderQR':
                if "uid" not in event['body-json']:
                    raise ValueError('Missing CardHolderUID')

                cardholder_data = api.get_card_holder(event['body-json']['uid'])
                badge_b64 = Badge(cardholder_data, app_config).makePNG()

                resp = {
                    "qr": badge_b64,
                    "cardCode": cardholder_data['cards'][0]['cardCode']
                }
                return resp

            if operation == 'getCardHolders':
                return handle_get_card_holders_request(api, event['body-json'])

            if operation == 'getCardHolderPhoto':
                if "uid" not in event['body-json']:
                    raise ValueError('Missing uid')
                card_holders = api.get_card_holder_photo(event['body-json']['uid'])
                return card_holders

            if operation == 'createCardHolder':
                cardholder_uid = handle_create_card_holder_request(api, event['body-json'])
                resp = {
                    "cardholderUID": cardholder_uid
                }
                return resp

            if operation == 'deleteCardHolder':
                if "uid" not in event['body-json']:
                    raise ValueError('Missing uid')
                api.delete_card_holder(event['body-json']['uid'])
                resp = {
                    "cardholderUID": event['body-json']['uid']
                }
                return resp

            if operation == 'getSecurityGroups':
                return api_security_groups.get_security_groups()

        raise ValueError("Operation not supported")
    #  AWS API-Gateway HTTP Status Code Mappings
    #  400    BadRequest.*
    #  401    Unauthorized.*
    #  500    InternalServerError.*
    #  503    ServiceUnavailable.*
    #  504    GatewayTimeout.*
    except ValueError as e:
        raise SensorVMSError("BadRequest: " + str(e)) from e
    except NotFoundError as e:
        raise SensorVMSError("NotFound: " + str(e)) from e
    except requests.exceptions.HTTPError as e:
        raise SensorVMSError("InternalServerError: " + str(e)) from e
    except requests.exceptions.ConnectionError as e:
        raise SensorVMSError("ServiceUnavailable: The GuardPoint Server is unreachable/offline." + str(e)) from e
    except requests.exceptions.Timeout as e:
        raise SensorVMSError("GatewayTimeout: The GuardPoint Server is taking too long to respond." + str(e)) from e
    except requests.exceptions.RequestException as e:
        raise SensorVMSError("InternalServerError: " + str(e)) from e
    except ServiceUnavailable as e:
        raise SensorVMSError("ServiceUnavailable: " + str(e)) from e
    except UnauthorizedError as e:
        raise SensorVMSError("Unauthorized: " + str(e)) from e
    except TokenAPIError as e:
        raise SensorVMSError("Unauthorized: " + str(e)) from e
    except Exception as e:
        raise SensorVMSError("InternalServerError: " + str(e)) from e


def handle_get_card_holders_request(api, event):
    if "searchPhrase" not in event:
        event['searchPhrase'] = None
    if "offset" not in event:
        event['offset'] = 0
    if "limit" not in event:
        event['limit'] = 10

    card_holders = api.get_card_holders(offset=event['offset'], limit=event['limit'],
                                        searchPhrase=event['searchPhrase'],
                                        mode=NameFilterMode.EITHER_OR)

    return card_holders


def handle_create_card_holder_request(api, event):
    # Perform User Input Validation
    if "firstName" not in event:
        raise ValueError('Missing firstName')
    if "lastName" not in event:
        raise ValueError('Missing lastName')
    if "company" not in event:
        raise ValueError('Missing company')
    if "email" not in event:
        raise ValueError('Missing Email')
    if "fromDateTimeValid" not in event:
        raise ValueError('Missing fromDateTimeValid')
    if "toDateTimeValid" not in event:
        raise ValueError('Missing toDateTimeValid')
    if "photo" not in event:
        raise ValueError('Missing photo')
    if "securityGroupUID" not in event:
        raise ValueError('Missing securityGroupUID')
    if "cardCode" not in event:
        event['cardCode'] = None
    else:
        if event['cardCode'] == '00000000':
            event['cardCode'] = None
        elif len(event['cardCode']) != 8:
            raise ValueError("BadRequest - Cardcode Length")
        elif not is_hex(event['cardCode']):
            raise ValueError("BadRequest - Cardcode contains invalid hexadecimal characters")
    if len(event['firstName']) < 1 or len(event['firstName']) > 100:
        raise ValueError('firstName length')
    if len(event['lastName']) < 1 or len(event['lastName']) > 100:
        raise ValueError('lastName length')
    if event['email']:
        if len(event['email']) > 0:
            if not validators.email(event['email']):
                raise ValueError(f'Invalid Email')
    if event['photo']:
        if len(event['photo']) < 100:
            event['photo'] = None
        elif len(event['photo']) > MAX_PHOTO_SIZE:
            raise ValueError('Photo too big')
        if not is_base64(event['photo']):
            raise ValueError('Photo encoding')
    try:
        from_datetime_valid = event["fromDateTimeValid"]
        from_datetime_valid = from_datetime_valid.replace('.000Z', '+00:00')
        from_datetime_valid = datetime.fromisoformat(from_datetime_valid)
    except Exception:
        raise ValueError(f'fromDateTimeValid Format({from_datetime_valid})')
    try:
        to_datetime_valid = event["toDateTimeValid"]
        to_datetime_valid = to_datetime_valid.replace('.000Z', '+00:00')
        to_datetime_valid = datetime.fromisoformat(to_datetime_valid)
    except Exception:
        raise ValueError(f'toDateTimeValid Format({to_datetime_valid})')

    # End Of Input Validation

    # Resize Photo
    # if event['photo']:
    #    event['photo'] = resize_image(event['photo'])

    card_holder_uid = api.create_card_holder_with_detail_and_card(
        first_name=event['firstName'],
        last_name=event['lastName'],
        company=event['company'],
        from_datetime_valid=from_datetime_valid.isoformat(),
        to_datetime_valid=to_datetime_valid.isoformat(),
        photo=event['photo'],
        email=event['email'],
        security_group_uid=event['securityGroupUID'],
        card_code=event['cardCode'])

    return card_holder_uid